 unit UartLaer; // v 02.01

interface
uses Windows;

const
 pinRTS  = DWORD($0001);
 pinDTR  = DWORD($0002);
 pinCTS  = DWORD($0010);
 pinDSR  = DWORD($0020);
 pinRING = DWORD($0040);
 pinRLSD = DWORD($0080);

type
 TUartEvent = (evZero, evOpen, evState, evRead, evReadT, evWrite, evClose, evBreak, evError);
 TUartOper  = (opZero, opClose, opPost, opState, opRead, opWrite);
 TUartState = (stZero, stToClose, stWork);

type TUartLaer=class
const
 cRxDefLen=1024*16;
 cRxDelay=250;
 cMaskEvent=(EV_DSR or EV_CTS or EV_RING or EV_RLSD or EV_RXFLAG or EV_RXCHAR);
type
 PUartIo=^TUartIo;
 TUartIo = record
  Overlapped :TOverlapped;
  bIo        :Cardinal;
  Oper       :TUartOper
 end;
 TUartProc = procedure(Obj:Pointer;Uart:TUartLaer;Event:TUartEvent);

 private
  fBuff   :PByte;
  fLen    :Cardinal;
  fMaxLen :Cardinal;
  fRxLen  :Cardinal;
  fTxLen  :Cardinal;
  fTxed   :Cardinal;
  fRxed   :Cardinal;

  fObj    :Pointer;
  fErr    :Cardinal;
  fCfg    :TCOMMCONFIG;
  fTmo    :TCOMMTIMEOUTS;
  fPin    :Cardinal;
  fStat   :Cardinal;
  fFree   :Boolean;
  fPort   :THandle;
  fProc   :TUartProc;
  fName   :string;
  fRead   :Boolean;
  fFlags  :Cardinal;
  fEvent  :THandle;
  fState  :TUartState;
  fBreak  :Boolean;

  ioPost,
  ioState,
  ioWrite,
  ioRead  :TUartIo;

  function  IsInit:Boolean;
  function  SetTmo(tmo:Cardinal;delay:Cardinal=cRxDelay): Boolean;
  procedure SetPin(UartPin:Cardinal);
  procedure SetCfg(Cfg:TCommConfig);
  procedure Event(Ev:TUartEvent);
  procedure Io(Io:PUartIo;bIo:Cardinal;IsOk:Boolean);
  procedure Read;
  procedure State;
  procedure InitCfg(pCfg:PCOMMCONFIG);
  procedure SetRead(f:Boolean);
  procedure _Close;

 public
  Tag    :Integer;

  property Buff: PByte    read fBuff;
  property Len:  Cardinal read fLen write fLen;
  property Err:  Cardinal read fErr;
  property RxLen:Cardinal read fRxLen;
  property TxLen:Cardinal read fTxLen;
  property Pin:  Cardinal read fPin write SetPin;
  property Cfg:  TCommConfig read fCfg write SetCfg;
  property AutoRead:Boolean read fRead write SetRead;

  property Rxed:Cardinal read fRxed;
  property Txed:Cardinal read fTxed;
  property Stat:Cardinal read fStat;

  function StrErr:string;

  constructor Create(Obj:Pointer;Proc:TUartProc;RxBuffLen:Cardinal=cRxDefLen);
  procedure   Free;

  procedure   Open(Name:String;pCfg:Pointer);
  procedure   Close;

  procedure   Write(Buff:PByte;Len:Cardinal);
  function    ReadT(Buff:PByte;Len:Cardinal;TimeOut:Cardinal;Delay:Cardinal=cRxDelay):Cardinal;

  function    GuiCfg(PortName:PChar):Boolean;

 private class var fCn: LongInt;
 private class var hCp: THandle;
 private class var hTs: array of THandle;
 public  class var MaxIoThreads:Cardinal;

 private class procedure Fin;
 private class procedure Init;
 private class procedure IoThread;
end;
{------------------------------------------------------------------------------}
implementation
{------------------------------------------------------------------------------}
{ TUartLaer }

procedure TUartLaer.SetCfg(Cfg: TCommConfig);
begin
 if SetCommConfig(fPort,Cfg,sizeof(Cfg)) then InitCfg(@Cfg)
 else begin
  InitCfg(nil);
  SetCommConfig(fPort,fCfg,sizeof(fCfg));
 end;
 SetPin(fPin);
end;

procedure TUartLaer.Close;
begin
 if fPort=0 then Exit;
 if fState=stToClose then Exit;
 fState:=stToClose;
 CloseHandle(fPort);
 if fBreak then _Close;
end;

constructor TUartLaer.Create(Obj:Pointer;Proc:TUartProc;RxBuffLen:Cardinal=cRxDefLen);
begin
 Inc(fCn);
 fObj:=Obj;
 @fProc:=@Proc;
 if RxBuffLen=0 then RxBuffLen:=cRxDefLen;
 fMaxLen:=RxBuffLen;
 fBuff:=GetMemory(fMaxLen);
 ioPost.Oper:=opPost;
 ioRead.Oper:=opRead;
 ioState.Oper:=opState;
 ioWrite.Oper:=opWrite;
end;

procedure TUartLaer.Event(Ev: TUartEvent);
var
 fR:Boolean;
begin
 fR:=True;
 case Ev of
  evOpen: ;
  evState: begin
            GetCommModemStatus(fPort,fStat);
           end;
  evRead:  if ioRead.bIo>0 then begin
            fRxLen:=ioRead.bIo;
            Inc(fRxed,fRxLen);
            fLen:=fLen+fRxLen;
           end else fR:=False;
  evReadT: ;
  evWrite: begin
            fTxLen:=ioWrite.bIo;
            Inc(fTxed,fTxLen);
           end;
  evClose: ;
  evBreak: ;
  evError: fErr:=GetLastError;
 end;
 if (fR)and(fObj<>nil)and(@fProc<>nil) then fProc(fObj,Self,Ev);
 case Ev of
  evOpen:  State;
  evState: State;
  evRead:  if fRead then State;
  evReadT: ;
  evWrite: ;
  evClose: _Close;
  evBreak: begin fBreak:=True; Close; end;
  evError: Close ;
 end;
end;

procedure TUartLaer.Free;
begin
 if fState<>stZero then begin
  if fState=stWork then begin
   fFree:=True;
   fObj:=nil;
   Close;
  end; Exit; end;
 Dec(fCn); if fCn=0 then Fin;
 inherited;
end;

function TUartLaer.GuiCfg(PortName:PChar):Boolean;
var
 Cfg:TCommConfig;
 Name:PChar;
begin
 Cfg:=fCfg;
 if Cfg.dwSize=0 then begin
  Cfg.dwSize:=sizeof(Cfg);
  Cfg.dcb.DCBlength:=sizeof(Cfg.dcb);
 end;
 if PortName<>nil then Name:=PortName
                  else Name:=PChar(fName);
 Result:=CommConfigDialog(Name,0,Cfg);
 if Result then SetCfg(Cfg)
           else fErr:=GetLastError;
end;

class procedure TUartLaer.Init;
var
 si:_SYSTEM_INFO;
 i :Integer;
 n :Cardinal;
begin
 if hCp<>0 then Exit;
 hCp:=CreateIoCompletionPort(INVALID_HANDLE_VALUE,0,0,0);
 if (hCp=0) then Exit;
 GetSystemInfo(si);
 n:=si.dwNumberOfProcessors;
 if MaxIoThreads=0 then MaxIoThreads:=1;
 if MaxIoThreads>0 then if n>MaxIoThreads then n:=MaxIoThreads;
 SetLength(hTs,n);
 for i:=n-1 downto 0 do
  hTs[i]:=BeginThread(nil,0,@TUartLaer.IoThread,nil,0,n);
end;

procedure TUartLaer.InitCfg(pCfg: PCOMMCONFIG);
begin
 if (pCfg=nil) then begin
  fCfg.dwSize:=       sizeof(fCfg);
  fCfg.dcb.DCBlength:=sizeof(fCfg.dcb);
  fCfg.dcb.BaudRate:= CBR_115200;
  fCfg.dcb.ByteSize:= 8;
  fCfg.dcb.Parity:=   NOPARITY;
  fCfg.dcb.StopBits:= ONESTOPBIT;
  fCfg.dcb.Flags:=    RTS_CONTROL_ENABLE or DTR_CONTROL_ENABLE;
 end else CopyMemory(@fCfg,pCfg,sizeof(fCfg));
end;

class procedure TUartLaer.IoThread;
var
 bIo  :Cardinal;
 pIo  :PUartIo;
 Obj  :TUartLaer;
 Key  :Cardinal absolute Obj;
 pOv  :POverlapped absolute pIo;
 R    :Boolean;
begin
 while True do try
  R:=GetQueuedCompletionStatus(hCp,bIo,Key,pOv,INFINITE);
  if (Obj<>nil)and(pIo<>nil) then Obj.Io(pIo,bIo,R)
  else if (Obj=nil)and(pIo=nil) then Break else begin
   if not R then Break; //to detailed debug
  end;
 except end;
end;

function TUartLaer.IsInit: Boolean;
begin
 Result:=hCp<>0;
 if not Result then begin
  Init;
  Result:=hCp<>0;
 end;
end;

procedure TUartLaer.Open(Name: String; pCfg: Pointer);
var
 err:Cardinal;
begin
 while True do begin
  if (not IsInit)or(fState<>stZero) then Break;
  fRead:=True;
  fName:=Name;
  fPort:=CreateFile(PChar('\\.\'+Name),GENERIC_READ or GENERIC_WRITE,0,nil,
                    OPEN_EXISTING,FILE_ATTRIBUTE_NORMAL or FILE_FLAG_OVERLAPPED,0);
  if fPort=INVALID_HANDLE_VALUE then begin fPort:=0; Break; end;
  while True do begin
   fEvent:=CreateEvent(nil,False,False,nil); if fEvent=0 then Break;
   if CreateIoCompletionPort(fPort,hCp,DWORD(Self),0)=0 then Break;
   InitCfg(pCfg);
   SetTmo(MAXDWORD);
   fLen:=0;
   fFree:=False;
   fBreak:=False;
   fState:=stWork;
   SetCfg(fCfg);
   SetPin(pinDTR or pinRTS);
   SetCommMask(fPort,cMaskEvent);
   Event(evOpen);
   Exit;
  end;
  err:=GetLastError;
  if fPort<>INVALID_HANDLE_VALUE then CloseHandle(fPort);
  fState:=stZero;
  SetLastError(err);
 end;
 Event(evError);
end;

procedure TUartLaer.Read;
var
 buff:PByte;
 mlen:Cardinal;
begin
 if fState<>stWork then Exit;
 buff:=PByte(DWORD(fBuff)+fLen);
 if fLen>=fMaxLen then fLen:=0;
 mlen:=fMaxLen-fLen;
 if (ReadFile(fPort,buff^,mlen,ioRead.bIo,@ioRead.Overlapped)=false)
  then if (GetLastError()<>ERROR_IO_PENDING) then Event(evError);
end;

function TUartLaer.ReadT(Buff:PByte;Len:Cardinal;TimeOut:Cardinal;Delay:Cardinal=cRxDelay):Cardinal;
var
 Io:TUartIo;
begin
 Result:=0;
 if fRead then Exit;
 if fState<>stWork then Exit;
 ResetEvent(fEvent);
 ZeroMemory(@Io,Sizeof(Io));
 Io.Oper:=opRead;
 Io.Overlapped.hEvent:=fEvent;
 SetTmo(TimeOut,Delay);
 while True do begin
  if not ReadFile(fPort,Buff^,Len,Io.bIo,@Io.Overlapped) then
   if GetLastError<>ERROR_IO_PENDING then Break;
  if not GetOverlappedResult(fPort,Io.Overlapped,Result,True) then Break;
  Exit;
 end;
 Event(evError);
end;

procedure TUartLaer.SetPin(UartPin: Cardinal);
begin
 if fState<>stWork then Exit;
 if (UartPin and pinRTS)xor(fPin and pinRTS)<>0 then
  if (UartPin and pinRTS)<>0 then EscapeCommFunction(fPort,SETRTS)
                             else EscapeCommFunction(fPort,CLRRTS);
 if (UartPin and pinDTR)xor(fPin and pinDTR)<>0 then
  if (UartPin and pinDTR)<>0 then EscapeCommFunction(fPort,SETDTR)
                             else EscapeCommFunction(fPort,CLRDTR);
 fPin:=UartPin;
end;

procedure TUartLaer.SetRead(f: Boolean);
begin
 if f=fRead then Exit;
 if f then SetTmo(MAXDWORD);
 fRead:=f;
end;

function TUartLaer.SetTmo(tmo:Cardinal;delay:Cardinal=cRxDelay): Boolean;
begin
 if tmo=MAXDWORD then begin
  fTmo.ReadTotalTimeoutMultiplier:=0;
  fTmo.ReadTotalTimeoutConstant:=Delay;
  fTmo.ReadIntervalTimeout:=MAXDWORD;
 end else begin
  fTmo.ReadTotalTimeoutMultiplier:=0;
  fTmo.ReadTotalTimeoutConstant:=tmo;
  fTmo.ReadIntervalTimeout:=Delay;
 end;
 Result:=SetCommTimeouts(fPort,fTmo);
end;

procedure TUartLaer.State;
begin
 if fState<>stWork then Exit;
 fFlags:=0;
 if (WaitCommEvent(fPort,fFlags,@ioState.Overlapped)=false) then
  if (GetLastError()<>ERROR_IO_PENDING) then Event(evError);
end;

class procedure TUartLaer.Fin;
var
 i:integer;
begin
 if hCp=0 then Exit;
 for i:=Length(hTs)-1 downto 0 do PostQueuedCompletionStatus(hCp,0,0,nil);
 WaitForMultipleObjects(Length(hTs),PWOHandleArray(hTs),True,1000);
 for i:=Length(hTs)-1 downto 0 do CloseHandle(hTs[i]);
 SetLength(hTs,0);
 CloseHandle(hCp);
 hCp:=0;
end;

function TUartLaer.StrErr: string;
const
 maxErrLook=256;
begin
 SetLength(Result,maxErrLook);
 FormatMessage(FORMAT_MESSAGE_FROM_SYSTEM,nil,fErr,
               LANG_USER_DEFAULT,PChar(Result),maxErrLook,nil);
 SetLength(Result,Length(PChar(Result))-2);
end;

procedure TUartLaer.Io(Io:PUartIo;bIo:Cardinal;IsOk:Boolean);
begin
 if IsOk then begin
  Io^.bIo:=bIo;
  if (bIo>0) then
   case Io^.Oper of
    opPost:;
    opRead:
      if fRead then Event(evRead);
    opWrite:
      Event(evWrite);
    opState: begin
     if ((fFlags and (EV_CTS or EV_DSR or EV_RING or EV_RLSD))<>0)
      then Event(evState);
     if ((fFlags and (EV_RXCHAR or EV_RXFLAG))<>0)
      then if fRead then Read();
     //if (fFlags=0)
      //then Event(evBreak);
    end;
   end
  else if Io^.Oper=opRead then
   begin if fRead then Event(evRead); end;
 end
 else begin
  fErr:=GetLastError();
  if fState=stToClose
   then Event(evClose)
   else Event(evBreak);
 end;
end;

procedure TUartLaer.Write(Buff: PByte; Len: Cardinal);
begin
 if (fState<>stWork)or(Len=0) then Exit;
 if (WriteFile(fPort,Buff^,Len,ioWrite.bIo,@ioWrite.Overlapped)=false)
  then if (GetLastError()<>ERROR_IO_PENDING) then Event(evError);
end;

procedure TUartLaer._Close;
begin
 if fState<>stToClose then Exit;
 CloseHandle(fEvent);
 fEvent:=0;
 fPort:=0;
 fState:=stZero;
 if fFree then Free;
end;

end.



